# Copyright (c) 2019, Moritz E. Beber.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a compartment ORM model."""


from typing import List, Optional

from sqlalchemy import Column, String
from sqlalchemy.orm import relationship

from .base import Base
from .compartment_annotation import CompartmentAnnotation
from .compartment_name import CompartmentName


class Compartment(Base):
    """
    Define a compartment ORM model.

    Attributes
    ----------
    notes : str, optional
    names : list of cobra_component_models.orm.CompoundName, optional
    annotation : list of cobra_component_models.orm.CompoundAnnotation, optional

    """

    __tablename__ = "compartments"

    notes: Optional[str] = Column(String, nullable=True)
    names: List[CompartmentName] = relationship("CompartmentName")
    annotation: List[CompartmentAnnotation] = relationship("CompartmentAnnotation")



# Copyright (c) 2019, Moritz E. Beber.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a custom pydantic type for component notes."""


class NotesType(str):
    """Define a custom pydantic type for component notes."""

    @classmethod
    def __get_validators__(cls):
        """Implement pydantic special behavior."""
        yield cls.validate

    @classmethod
    def validate(cls, value: str):
        """Validate the notes value."""
        if not isinstance(value, str):
            raise ValueError(f"Notes need to be of type str not {type(value)}.")
        # TODO (Moritz Beber): Check XML. See also
        #  SBML spec section 3.2.5 Notes.
        # elif False:
        #     raise ValueError("SBML notes need to be well-formed XHTML content.")
        return value



# Copyright (c) 2019, Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a mixin that tracks creation and update timestamps."""


from datetime import datetime, timezone

from sqlalchemy import Column, DateTime


def timezone_aware_now():
    """Return the date and time in this moment in the universal timezone."""
    return datetime.now(timezone.utc)


class TimestampMixin:
    """
    Define creation and update time columns to be mixed in with other tables.

    Attributes
    ----------
    created_on : datetime
        By default this value is populated at instantiation with the time of
        the moment.
    updated_on : datetime
        The time is automatically populated whenever the database model is
        updated.

    """

    created_on: datetime = Column(
        DateTime(timezone=True), nullable=False, default=timezone_aware_now
    )
    updated_on: datetime = Column(
        DateTime(timezone=True), nullable=True, onupdate=timezone_aware_now
    )



# The MIT License (MIT)
#
# Copyright (c) 2019, Moritz E. Beber.
# Copyright (c) 2018 Institute for Molecular Systems Biology, ETH Zurich.
# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


"""Provide a compound ORM model."""


from typing import List, Optional

from sqlalchemy import Column, Float, String
from sqlalchemy.orm import relationship

from . import CompoundAnnotation, CompoundName
from .base import Base


class Compound(Base):
    """
    Define a compound ORM model.

    Attributes
    ----------
    inchi : str, optional
    inchi_key : str, optional
    smiles : str, optional
    charge : float, optional
    chemical_formula : str, optional
    notes : str, optional
    names : list of cobra_component_models.orm.CompoundName, optional
    annotation : list of cobra_component_models.orm.CompoundAnnotation, optional

    """

    __tablename__ = "compounds"

    inchi: Optional[str] = Column(String, nullable=True, index=True)
    inchi_key: Optional[str] = Column(String(27), nullable=True, index=True)
    smiles: Optional[str] = Column(String, nullable=True, index=True)
    charge: Optional[float] = Column(Float, nullable=True)
    chemical_formula: Optional[str] = Column(String, nullable=True, index=True)
    notes: Optional[str] = Column(String, nullable=True)
    names: List[CompoundName] = relationship("CompoundName")
    annotation: List[CompoundAnnotation] = relationship("CompoundAnnotation")

    def __repr__(self):
        """Return a string representation of the object."""
        return f"{type(self).__name__}(id={self.id}, inchi_key={self.inchi_key})"



# Copyright (c) 2019, Moritz E. Beber.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a custom pydantic type for component annotations."""


from importlib.resources import open_text
from typing import List, Tuple

from ... import data


# A list of qualifiers taken from https://co.mbine.org/standards/qualifiers.
with open_text(data, "biology_qualifiers.txt") as handler:
    BIOLOGY_QUALIFIERS = frozenset(l.strip() for l in handler.readlines())


class AnnotationType(tuple):
    """
    Define a custome pydantic type for component annotations.

    An annotation consists of a pair of strings representing a biology qualifier and
    an identifier from a specific namespace.

    """

    @classmethod
    def __get_validators__(cls):
        """Implement pydantic special behavior."""
        yield cls.validate

    @classmethod
    def validate(cls, value: List[str]) -> Tuple[str, str]:
        """Validate and transform the given annotation."""
        try:
            qualifier, identifier = value
        except (ValueError, TypeError) as exc:
            raise ValueError(
                "Each annotation must consist of a biology qualifier "
                "(https://co.mbine.org/standards/qualifiers), "
                "identifier pair."
            ) from exc
        if qualifier not in BIOLOGY_QUALIFIERS:
            raise ValueError(
                "The qualifier must be one of the valid biology qualifiers defined "
                "at https://co.mbine.org/standards/qualifiers."
            )
        return qualifier, identifier



# Copyright (c) 2019, Moritz E. Beber.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a biology qualifier ORM model."""


from __future__ import annotations

from importlib.resources import open_text
from typing import Dict

from sqlalchemy import Column, String, exists

from .. import data
from .base import Base


class BiologyQualifier(Base):
    """
    Define a BioModels biology qualifier ORM model.

    You can read more about them at http://co.mbine.org/standards/qualifiers.

    Attributes
    ----------
    qualifier : str
        The text value of the qualifier.

    """

    __tablename__ = "biology_qualifiers"

    qualifier: str = Column(String, nullable=False, index=True, unique=True)

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f"{type(self).__name__}(qualifier={self.qualifier})"

    @classmethod
    def load(cls, session):
        """Load all known biology qualifiers into the given database."""
        with open_text(data, "biology_qualifiers.txt") as handler:
            qualifiers = [l.strip() for l in handler.readlines()]
        for qual in qualifiers:
            if (
                qual
                and not session.query(exists().where(cls.qualifier == qual)).scalar()
            ):
                session.add(cls(qualifier=qual))
        session.commit()

    @classmethod
    def get_map(cls, session) -> Dict[str, BiologyQualifier]:
        """Extract a mapping from biology qualifiers to ORM instances."""
        return {qualifier.qualifier: qualifier for qualifier in session.query(cls)}



# The MIT License (MIT)
#
# Copyright (c) 2019, Moritz E. Beber.
# Copyright (c) 2018 Institute for Molecular Systems Biology, ETH Zurich.
# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


"""Provide a MIRIAM compliant Identifiers.org namespace ORM model."""


from __future__ import annotations

import re
from typing import ClassVar, Dict, Optional, Pattern

from sqlalchemy import Boolean, Column, String
from sqlalchemy.orm import reconstructor, validates

from .base import Base


class Namespace(Base):
    r"""
    Define a MIRIAM compliant Identifiers.org namespace ORM model.

    Attributes
    ----------
    miriam_id : str
        The MIRIAM namespace identifier for itself, e.g., MIR:00000567.
    prefix : str
        The MIRIAM namespace prefix, e.g., 'metanetx.chemical'.
    pattern : str
        The regular expression pattern to validate against identifiers in the
        namespace, e.g., ``^(MNXM\d+|BIOMASS)$``.
    embedded_prefix : bool
        Whether or not identifiers of this namespace have an embedded `prefix`, e.g.,
        'CHEBI:52971'.
    name : str, optional
        The namespace's common name.
    description : str, optional
        A short description of the namespace.

    """

    __tablename__ = "namespaces"

    miriam_id: str = Column(String(12), nullable=False, index=True, unique=True)
    # The currently longest prefix on Identifiers.org is 22.
    prefix: str = Column(String(22), nullable=False, index=True, unique=True)
    pattern: str = Column(String, nullable=False)
    embedded_prefix: bool = Column(Boolean, default=False, nullable=False)
    name: Optional[str] = Column(String, nullable=True)
    description: Optional[str] = Column(String, nullable=True)

    # Define normal Python class variables.
    _identifier_pattern: ClassVar[Pattern] = re.compile(r"^MIR:\d{8}$")

    def __init__(self, **kwargs):
        """
        Initialize a namespace object.

        While defining an init method is usually not necessary for SQLAlchemy
        declarative models, we do it here to insert the `compiled_pattern` attribute.

        """
        super().__init__(**kwargs)
        self.compiled_pattern = re.compile(self.pattern)

    def __repr__(self):
        """Return a string representation of the object."""
        return f"{type(self).__name__}(prefix={self.prefix})"

    @reconstructor
    def init_on_load(self):
        """Compile the identifier pattern on load from database."""
        self.compiled_pattern = re.compile(self.pattern)

    @validates("miriam_id")
    def validate_identifier(self, _, miriam_id: str) -> str:
        """Validate the MIRIAM identifier against the pattern."""
        if self._identifier_pattern.match(miriam_id) is None:
            raise ValueError(
                f"The namespace's identifier '{miriam_id}' does not match the "
                f"official pattern '^MIR:\\d{8}$'."
            )
        return miriam_id

    @classmethod
    def get_map(cls, session) -> Dict[str, Namespace]:
        """Extract a mapping from namespace prefix to ORM instances."""
        return {ns.prefix: ns for ns in session.query(cls)}



# Copyright (c) 2019, Moritz E. Beber.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a reaction annotation ORM model."""


from sqlalchemy import Column, ForeignKey, Integer

from .base import Base
from .mixin import AnnotationMixin


class ReactionAnnotation(AnnotationMixin, Base):
    """
    Define a reaction annotation ORM model.

    Attributes
    ----------
    reaction_id : int
        The reaction being annotated.

    """

    __tablename__ = "reaction_annotations"

    reaction_id: int = Column(Integer, ForeignKey("reactions.id"), nullable=False)

    def __repr__(self):
        """Return a string representation of the object."""
        return (
            f"{type(self).__name__}(reaction={self.reaction_id}, "
            f"identifier={self.identifier}, "
            f"namespace={self.namespace_id}, "
            f"qualifier={self.qualifier_id})"
        )



# Copyright (c) 2019, Moritz E. Beber.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a compartment name ORM model."""


from sqlalchemy import Column, ForeignKey, Integer

from .base import Base
from .mixin import NameMixin


class CompartmentName(NameMixin, Base):
    """
    Define a compartment name ORM model.

    Attributes
    ----------
    compartment_id : int
        The compartment being named.

    """

    __tablename__ = "compartment_names"

    compartment_id: int = Column(Integer, ForeignKey("compartments.id"), nullable=False)

    def __repr__(self):
        """Return a string representation of the object."""
        return (
            f"{type(self).__name__}(compartment={self.compartment_id}, "
            f"name={self.name}, "
            f"namespace={self.namespace_id})"
        )



# Copyright (c) 2019, Moritz E. Beber.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a reaction serializer."""


from typing import Dict, List, Optional, Tuple

from ..io import ParticipantModel, ReactionModel
from ..orm import (
    Compartment,
    Compound,
    Participant,
    Reaction,
    ReactionAnnotation,
    ReactionName,
)
from .abstract_serializer import AbstractSerializer


class ReactionSerializer(AbstractSerializer):
    """Define a reaction serializer."""

    def __init__(
        self,
        compartment2id: Optional[Dict[Compartment, str]] = None,
        compound2id: Optional[Dict[Compound, str]] = None,
        id2compartment: Optional[Dict[str, Compartment]] = None,
        id2compound: Optional[Dict[str, Compound]] = None,
        **kwargs
    ):
        """
        Initialize a reaction serializer.

        Parameters
        ----------
        compartment2id : dict
            A map from compartment database instances to string identifiers for them.
            Needed for serialization only.
        compound2id : dict
            A map from compound database instances to string identifiers for them.
            Needed for serialization only.
        id2compartment : dict
            A map from string identifiers to compartment database instances.
            Needed for deserialization only.
        id2compound : dict
            A map from string identifiers to compound database instances.
            Needed for deserialization only.

        Other Parameters
        ----------------
        kwargs
            Passed on to super class init method.

        """
        super().__init__(**kwargs)
        self.compartment2id = {} if compartment2id is None else compartment2id
        self.compound2id = {} if compound2id is None else compound2id
        self.id2compartment = {} if id2compartment is None else id2compartment
        self.id2compound = {} if id2compound is None else id2compound

    def serialize(self, component: Reaction) -> ReactionModel:
        """
        Serialize a reaction ORM model to a pydantic data model.

        Parameters
        ----------
        component : cobra_component_models.orm.Reaction
            The reaction ORM model instance to be serialized.

        Returns
        -------
        cobra_component_models.io.ReactionModel
            A corresponding pydantic reaction data model.

        Warnings
        --------
        Please ensure that all relationships of the reaction object have been eagerly
        loaded in order to avoid :math:`N+1` [R1]_ problems that may easily occur here.

        References
        ----------
        .. [R1] https://docs.sqlalchemy.org/en/13/glossary.html#term-n-plus-one-problem

        """
        names = self.serialize_names(component.names)
        annotation = self.serialize_annotation(component.annotation)
        reactants, products = self.serialize_participants(component.participants)
        return ReactionModel(
            id=str(component.id),
            notes=component.notes,
            names=names,
            annotation=annotation,
            reactants=reactants,
            products=products,
        )

    def serialize_participants(
        self, participants: List[Participant]
    ) -> Tuple[Dict[str, ParticipantModel], Dict[str, ParticipantModel]]:
        """Serialize the reactants and products."""
        reactants = {}
        products = {}
        for part in participants:
            if part.is_product:
                products[self.compound2id[part.compound]] = ParticipantModel(
                    stoichiometry=part.stoichiometry,
                    compartment=self.compartment2id[part.compartment],
                )
            else:
                reactants[self.compound2id[part.compound]] = ParticipantModel(
                    stoichiometry=part.stoichiometry,
                    compartment=self.compartment2id[part.compartment],
                )
        return reactants, products

    def deserialize(self, component_model: ReactionModel) -> Reaction:
        """
        Deserialize a pydantic reaction data model to an ORM model.

        Parameters
        ----------
        component_model : cobra_component_models.io.ReactionModel
            The pydantic reaction data model instance to be deserialized.

        Returns
        -------
        cobra_component_models.orm.Reaction
            A corresponding reaction ORM model.

        """
        reaction = Reaction(notes=component_model.notes)
        reaction.names.extend(
            self.deserialize_names(component_model.names, ReactionName)
        )
        reaction.annotation.extend(
            self.deserialize_annotation(component_model.annotation, ReactionAnnotation)
        )
        reaction.participants.extend(
            self.deserialize_participants(
                reactants=component_model.reactants, products=component_model.products
            )
        )
        return reaction

    def deserialize_participants(
        self,
        reactants: Dict[str, ParticipantModel],
        products: Dict[str, ParticipantModel],
    ) -> List[Participant]:
        """Deserialize the reactants and products."""
        participants = []
        for compound_id, part in reactants.items():
            participants.append(
                Participant(
                    compound=self.id2compound[compound_id],
                    compartment=self.id2compartment[part.compartment],
                    stoichiometry=part.stoichiometry,
                    is_product=False,
                )
            )
        for compound_id, part in products.items():
            participants.append(
                Participant(
                    compound=self.id2compound[compound_id],
                    compartment=self.id2compartment[part.compartment],
                    stoichiometry=part.stoichiometry,
                    is_product=True,
                )
            )
        return participants



# Copyright (c) 2019, Moritz E. Beber.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a reaction participant (reagent/product) ORM model."""


from typing import Optional

from sqlalchemy import Boolean, Column, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

from .base import Base
from .compartment import Compartment
from .compound import Compound


class Participant(Base):
    """
    Define a reaction participant (reagent/product) ORM model.

    Attributes
    ----------
    stoichiometry : str

    """

    __tablename__ = "participants"

    reaction_id: int = Column(Integer, ForeignKey("reactions.id"), nullable=False)
    compound_id: int = Column(Integer, ForeignKey("compounds.id"), nullable=False)
    compound: Compound = relationship("Compound")
    stoichiometry: str = Column(String, nullable=False)
    is_product: bool = Column(Boolean, nullable=False)
    compartment_id: Optional[int] = Column(
        Integer, ForeignKey("compartments.id"), nullable=True
    )
    compartment: Compartment = relationship("Compartment")

    def __repr__(self):
        """Return a string representation of the object."""
        return (
            f"{type(self).__name__}(id={self.id}, stoichiometry={self.stoichiometry})"
        )



# Copyright (c) 2019, Moritz E. Beber.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a compound serializer."""


from ..io import CompoundModel
from ..orm import Compound, CompoundAnnotation, CompoundName
from .abstract_serializer import AbstractSerializer


class CompoundSerializer(AbstractSerializer):
    """Define a compound serializer."""

    def __init__(self, **kwargs):
        """Initialize a compound serializer."""
        super().__init__(**kwargs)

    def serialize(self, component: Compound) -> CompoundModel:
        """
        Serialize a compound ORM model to a pydantic data model.

        Parameters
        ----------
        component : cobra_component_models.orm.Compound
            The compound ORM model instance to be serialized.

        Returns
        -------
        cobra_component_models.io.CompoundModel
            A corresponding pydantic compound data model.

        Warnings
        --------
        Please ensure that all relationships of the compound object have been eagerly
        loaded in order to avoid :math:`N+1` [C1]_ problems that may easily occur here.

        References
        ----------
        .. [C1] https://docs.sqlalchemy.org/en/13/glossary.html#term-n-plus-one-problem

        """
        names = self.serialize_names(component.names)
        annotation = self.serialize_annotation(component.annotation)
        annotation["inchi"] = [("is", component.inchi)]
        annotation["inchikey"] = [("is", component.inchi_key)]
        # SMILES are not yet Identifiers.org conform.
        annotation["smiles"] = [("is", component.smiles)]
        return CompoundModel(
            id=str(component.id),
            notes=component.notes,
            charge=component.charge,
            chemicalFormula=component.chemical_formula,
            names=names,
            annotation=annotation,
        )

    def deserialize(self, component_model: CompoundModel) -> Compound:
        """
        Deserialize a pydantic compound data model to an ORM model.

        Parameters
        ----------
        component_model : cobra_component_models.io.CompoundModel
            The pydantic compound data model instance to be deserialized.

        Returns
        -------
        cobra_component_models.orm.Compound
            A corresponding compound ORM model.

        """
        compound = Compound(
            charge=component_model.charge,
            chemical_formula=component_model.chemical_formula,
            notes=component_model.notes,
        )
        for structure in ["inchi", "smiles"]:
            if structure in component_model.annotation:
                # Set the structure and remove it from the annotations for later use.
                setattr(
                    compound, structure, component_model.annotation.pop(structure)[0][1]
                )
        if "inchikey" in component_model.annotation:
            compound.inchi_key = component_model.annotation.pop("inchikey")[0][1]
        compound.names.extend(
            self.deserialize_names(component_model.names, CompoundName)
        )
        compound.annotation.extend(
            self.deserialize_annotation(component_model.annotation, CompoundAnnotation)
        )
        return compound



# The MIT License (MIT)
#
# Copyright (c) 2018-2019, Moritz E. Beber.
# Copyright (c) 2018-2019, Institute for Molecular Systems Biology, ETH Zurich.
# Copyright (c) 2018-2019, Novo Nordisk Foundation Center for Biosustainability,
#     Technical University of Denmark.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


"""Provide the declarative base class for ORM models."""


from sqlalchemy import Column, Integer
from sqlalchemy.ext.declarative import declarative_base

from .mixin import TimestampMixin


class CustomBase(TimestampMixin):
    """
    Define a declarative base class with timestamps for ORM models.

    Attributes
    ----------
    id : int
        The auto-incrementing, integer, primary key in the table.

    """

    id = Column(Integer, primary_key=True, autoincrement=True)

    def __repr__(self):
        """Return a string representation of the object."""
        return f"{type(self).__name__}(id={self.id})"


Base = declarative_base(cls=CustomBase)



# Copyright (c) 2019, Moritz E. Beber.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a component annotation mixin with the corresponding ORM columns."""


import logging

from sqlalchemy import Column, ForeignKey, Integer, String
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship


logger = logging.getLogger(__name__)


class AnnotationMixin:
    """
    Define a component annotation mixin.

    Attributes
    ----------
    identifier : str
        The annotation identifier.
    namespace_id : int
        The foreign key of the related Identifiers.org namespace.
    qualifier_id : int
        The foreign key of the related biology qualifier.

    """

    identifier: str = Column(String, nullable=False, index=True)

    @declared_attr
    def namespace_id(cls):
        """Defer the namespace id field instantiation."""
        return Column(Integer, ForeignKey("namespaces.id"), nullable=False)

    @declared_attr
    def namespace(cls):
        """Defer the namespace field instantiation."""
        return relationship("Namespace")

    @declared_attr
    def biology_qualifier_id(cls):
        """Defer the biology qualifier id field instantiation."""
        return Column(Integer, ForeignKey("biology_qualifiers.id"), nullable=False)

    @declared_attr
    def biology_qualifier(cls):
        """Defer the biology qualifier field instantiation."""
        return relationship("BiologyQualifier")

    # FIXME: The namespace attribute is not initialized at the time of validation.
    #  Try to use events maybe
    #  (https://docs.sqlalchemy.org/en/13/orm/events.html#attribute-events)?
    # @validates("identifier")
    # def validate_identifier(self, _, identifier: str):
    #     """Use the namespace's pattern to validate the identifier."""
    #     if self.namespace.compiled_pattern.match(identifier) is None:
    #         logger.warning(
    #             f"Identifier '{identifier}' does not match "
    #             f"{self.namespace.name}'s pattern '{self.namespace.pattern}'."
    #         )
    #     return identifier



# Copyright (c) 2019, Moritz E. Beber.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a compartment serializer."""


from ..io import CompartmentModel
from ..orm import Compartment, CompartmentAnnotation, CompartmentName
from .abstract_serializer import AbstractSerializer


class CompartmentSerializer(AbstractSerializer):
    """Define a compartment serializer."""

    def __init__(self, **kwargs):
        """Initialize a compartment serializer."""
        super().__init__(**kwargs)

    def serialize(self, component: Compartment) -> CompartmentModel:
        """
        Serialize a compartment ORM model to a pydantic data model.

        Parameters
        ----------
        component : cobra_component_models.orm.Compartment
            The compartment ORM model instance to be serialized.

        Returns
        -------
        cobra_component_models.io.CompartmentModel
            A corresponding pydantic compartment data model.

        Warnings
        --------
        Please ensure that all relationships of the compartment object have been eagerly
        loaded in order to avoid :math:`N+1` [CC1]_ problems that may easily occur here.

        References
        ----------
        .. [CC1] https://docs.sqlalchemy.org/en/13/glossary.html#term-n-plus-one-problem

        """
        names = self.serialize_names(component.names)
        annotation = self.serialize_annotation(component.annotation)
        return CompartmentModel(
            id=str(component.id),
            notes=component.notes,
            names=names,
            annotation=annotation,
        )

    def deserialize(self, component_model: CompartmentModel) -> Compartment:
        """
        Deserialize a pydantic compartment data model to an ORM model.

        Parameters
        ----------
        component_model : cobra_component_models.io.CompartmentModel
            The pydantic compartment data model instance to be deserialized.

        Returns
        -------
        cobra_component_models.orm.Compartment
            A corresponding compartment ORM model.

        """
        compartment = Compartment(notes=component_model.notes)
        compartment.names.extend(
            self.deserialize_names(component_model.names, CompartmentName)
        )
        compartment.annotation.extend(
            self.deserialize_annotation(
                component_model.annotation, CompartmentAnnotation
            )
        )
        return compartment



# Copyright (c) 2019, Moritz E. Beber.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide SQLAlchemy ORM models for storing components."""


from .base import Base
from .biology_qualifier import BiologyQualifier
from .namespace import Namespace
from .compartment_annotation import CompartmentAnnotation
from .compartment_name import CompartmentName
from .compartment import Compartment
from .compound_annotation import CompoundAnnotation
from .compound_name import CompoundName
from .compound import Compound
from .reaction_annotation import ReactionAnnotation
from .reaction_name import ReactionName
from .participant import Participant
from .reaction import Reaction
from .abstract_component import (
    AbstractComponentAnnotation,
    AbstractComponentName,
    AbstractComponent,
)



# Copyright (c) 2019, Moritz E. Beber.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide the pydantic SBase data model."""


from typing import Dict, List, Optional

from pydantic import BaseModel, Schema

from .type import AnnotationType, NotesType


class SBaseModel(BaseModel):
    """Define the SBase data model."""

    id: Optional[str] = None
    sbo_term: Optional[str] = Schema(None, alias="sboTerm")
    notes: Optional[NotesType] = None
    names: Dict[str, List[str]] = {}
    annotation: Dict[str, List[AnnotationType]] = {}

    class Config:
        """Configure the SBase data model."""

        orm_mode = True



# Copyright (c) 2019, Moritz E. Beber.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide an abstract serializer."""


from abc import ABC, abstractmethod
from typing import Dict, List, Type

from ..io import SBaseModel
from ..io.type import AnnotationType
from ..orm import (
    AbstractComponent,
    AbstractComponentAnnotation,
    AbstractComponentName,
    BiologyQualifier,
    Namespace,
)


class AbstractSerializer(ABC):
    """Define an abstract serializer."""

    def __init__(
        self,
        biology_qualifiers: Dict[str, BiologyQualifier],
        namespaces: Dict[str, Namespace],
        **kwargs
    ):
        """
        Initialize the abstract base serializer.

        Parameters
        ----------
        biology_qualifiers : dict
            A mapping from biology qualifiers to their database instances.
        namespaces : dict
            A mapping from namespace prefixes to their database instances.

        Other Parameters
        ----------------
        kwargs
            Passed on to super class init method.

        """
        super().__init__(**kwargs)
        self.biology_qualifiers = biology_qualifiers
        self.namespaces = namespaces

    @abstractmethod
    def serialize(self, component: AbstractComponent) -> SBaseModel:
        """Serialize an ORM model to a pydantic data model."""
        pass

    def serialize_names(
        self, names: List[AbstractComponentName]
    ) -> Dict[str, List[str]]:
        """Serialize the component names."""
        obj = {}
        for name_model in names:
            obj.setdefault(name_model.namespace.prefix, []).append(name_model.name)
        return obj

    def serialize_annotation(
        self, annotation: List[AbstractComponentAnnotation]
    ) -> Dict[str, List[AnnotationType]]:
        """Serialize the component annotation."""
        obj = {}
        for ann in annotation:
            obj.setdefault(ann.namespace.prefix, []).append(
                (ann.biology_qualifier.qualifier, ann.identifier)
            )
        return obj

    @abstractmethod
    def deserialize(self, component_model: SBaseModel) -> AbstractComponent:
        """Deserialize a pydantic data model to an ORM model."""
        pass

    def deserialize_names(
        self, names_data: Dict[str, List[str]], orm_class: Type[AbstractComponentName]
    ) -> List[AbstractComponentName]:
        """Deserialize the component names."""
        result = []
        for prefix, names in names_data.items():
            namespace = self.namespaces[prefix]
            for name in names:
                result.append(orm_class(name=name, namespace=namespace))
        return result

    def deserialize_annotation(
        self,
        annotation_data: Dict[str, List[AnnotationType]],
        orm_class: Type[AbstractComponentAnnotation],
    ) -> List[AbstractComponentAnnotation]:
        """Deserialize the component annotation."""
        result = []
        for prefix, annotations in annotation_data.items():
            namespace = self.namespaces[prefix]
            for ann in annotations:
                qualifier = self.biology_qualifiers[ann[0]]
                result.append(
                    orm_class(
                        identifier=ann[1],
                        biology_qualifier=qualifier,
                        namespace=namespace,
                    )
                )
        return result



# Copyright (c) 2019, Moritz E. Beber.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a component name mixin with the corresponding ORM columns."""


from sqlalchemy import Column, ForeignKey, Integer, String
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship


class NameMixin:
    """
    Define a component name mixin.

    Attributes
    ----------
    name : str
        A common name for a component.
    namespace_id : int
        The foreign key of the related Identifiers.org namespace.

    """

    name: str = Column(String, nullable=False, index=True)

    @declared_attr
    def namespace_id(cls):
        """Defer the namespace id field instantiation."""
        return Column(Integer, ForeignKey("namespaces.id"), nullable=False)

    @declared_attr
    def namespace(cls):
        """Defer the namespace field instantiation."""
        return relationship("Namespace")
